# SimVascular Automating Model Generation For Image-Based Cardiac Flow Simulation

This is the source code for paper:

Kong, F., and Shadden, S. C. (August 7, 2020). "Automating Model Generation for Imagebased Cardiac Flow Simulation." ASME. J Biomech Eng. doi: https://doi.org/10.1115/1.4048032

The code consists of two parts

* Deep-learning based automatic segmentation of 3D CT/MR image data using an ensemble of 2D UNets
* Down-stream automatic model generation from segmentations for LV CFD simulations 

## Dependencies

* Segmentation and Model Generation
    * SimVascular
* Model Registration
    * [SimpleElastix](https://github.com/SuperElastix/SimpleElastix) (Registration)<sup>1</sup>
       * SimpleITK 
       <sup>1</sup>  Install SimpleElastix commit 8244e0001f4137514b0f545f1e846910b3dd7769.

## Segmentation Usage 

The segmentation models can generate segmentations for LV blood pool, LV myocardium, LA, RA, RV blood pool, aorta and pulmonary artery.
### Input Requirements
The preferred input format of the image volumes is **.nii.gz or nii**. VTK image volumes (.vti) are also accepted; however they should be reoriented to have an orientation matrix of identity. This is because the segmnetation method requires identity-oriented image volumes while the version of VTK within SimVascular does not include orientation matrix with VTI images.
The directory containing the input image data should be organized as follows:
```
image_dir
     |__ patient_id (optional)
         |__ image_volume0.nii.gz
         |__ image_volume1.nii.gz
         |__ image_volume2.nii.gz
         |__ ...
```
### Trained Models
We used the image and ground truth data provided by [MMWHS](http://www.sdspeople.fudan.edu.cn/zhuangxiahai/0/mmwhs/) to train our models. 
Our segmentation models were trainined simultaneously on CT and MR data and trained weights are [here](https://drive.google.com/open?id=162Xr5OezSZL-0K3aoYO7WnHWuGTEXkkj). 

### Prediction
To generate segmentations for 3D CT or MR image volumes:
```
from sv_auto_lv_modeling.auto_lv import Segmentation
data_path='/path/to/WS01_data'
seg = Segmentation()
seg.set_modality('ct')
seg.set_patient_id ('WS01')
seg.set_image_directory (data_path+'/01-Images')
seg.set_output_directory (data_path+'/02-Segmnts')
seg.set_model_directory ([data_path+'/Weights'])
seg.set_view ([2])
seg.generate_segmentation()
```
## LV Modeling Usage

The model construction pipeline takes in the generated segmentation and output reconstructed LV surface meshes for CFD simulations. The pipeline consists of the following four steps: 1) Construct LV surface meshes from segmentation results; 2) Register the surface meshes to get consistent mesh topology; 3) Obtain volumetric mesh using SimVascular; 4) Interpolate the registered surface meshes to obtain sufficient temporal resolution.

### 1.  Construct LV Surface Meshes with Tagged Boundary Faces
```
from sv_auto_lv_modeling.auto_lv import Modeling
data_path='/path/to/WS01_data'
surf = Modeling()
surf.set_segmentation_directory(data_path+'/02-Segmnts/WS01')
surf.set_output_directory (data_path+'/03-Surfaces/WS01')
surf.set_max_edge_size (3.5)
surf.generate_lv_modes ()
```

### 2.  Construct Point Corresponded LV Meshes from 4D Images
Building point-corresponded LV meshes require segmentations from all time frames. One surface mesh will be created at one time frame and propagated to the others by registering the corresponding segmentations. 
* Update `surface_registration.sh` with correct file and folder names. Specify the time phase id to construct LV surface mesh.

### 3.  Volumetric Meshing using SimVascular 
```
from sv_auto_lv_modeling.auto_lv import VolumeMesh 
data_path='/path/to/WS01_data'
vol = VolumeMesh()
vol.set_output_directory (data_path+'/05-VolMesh/WS01')
vol.set_max_edge_size (3.5)
vol.set_surface_model_filename (data_path+'/04-SurfReg/WS01/WS01_0.vti.vtp')
vol.generate_volume_mesh()
```



### 4.  Generate Mesh Motion File for [svFSI](https://github.com/SimVascular/svFSI)
* Input: Surface mesh from segmented geometry with the same connectivity.
  Output: Displacement files for all the surfaces in this [format](https://simvascular.github.io/docssvFSI.html#app_app_prescribed_wall_motion).

```
from sv_auto_lv_modeling.auto_lv import Interpolation
data_path='/path/to/WS01_data'
patient_id='WS01'
interp = Interpolation()
interp.set_surface_directory(data_path+'/04-SurfReg/'+patient_id)
interp.set_surface_phase_id(0)
interp.set_output_dir(data_path+'/05-VolMesh/'+patient_id)
interp.set_num_of_interpolations(99)
interp.set_num_of_cycles(1)
interp.set_cycle_length(1.25)
interp.interpolate()
```



## Acknowledgement
This work was supported by the NSF, Award #1663747. 

